#! /usr/bin/env perl
# version 1.2.6
########################## Copyright Notice ###################################
#            Copyright (c)1996 Universities Research Association, Inc.        #
#                        All Rights Reserved                                  #
###############################################################################
#
##################### Government Sponsorship Notice ###########################
#This material resulted from work developed under a Government Contract and   #
#is subject to the following license:  The Government retains a paid-up,      #
#nonexclusive, irrevocable worldwide license to reproduce, prepare derivative #
#works, perform publicly and display publicly by or for the Government,       #
#including the right to distribute to other Government contractors.  Neither  #
#the United States nor the United States Department of Energy, nor any of     #
#their employees, makes any warrenty, express or implied, or assumes any      #
#legal liability or responsibility for the accuracy, completeness, or         #
#usefulness of any information, apparatus, product, or process disclosed, or  #
#represents that its use would not infringe privately owned rights.           #
################################################################################
#
#  changes in version 1.2.6 - Benjamin Vial <benjamin.vial@urec.cnrs.fr>
# - parameters configuration using a centrally-managed configuration downloaded
#   by glite-netmon-config-update

#  changes in version 1.2.2 - Benjamin Vial <benjamin.vial@urec.cnrs.fr>
# - pinger sends an RTT measurement to the FIFO only if the pings were successful
#
# changes in version 1.1.5 - Paul Mealor <pdm@hep.ucl.ac.uk>
# - not only logs to file, but also attempts to talk to the glite-netmon-rgma
#   daemon which will publish data into RGMA
#
# changes in version 1.1.4 - XIWT
#    shortened some long lines; changed some locals to mys
#    reduce max process count from 15 to 5
#    collect seq nos and ping times and log them
#    replace multiple prints in log code with single print
#    fixed min/max/avg code for Linux (which reports floating points)
#    --jeremy@cnri.reston.va.us
#
# changes in version 1.1.3
# -fixed problem in generating correct command (Thanks to Mike Procario, CMU)
# 
# changes in version 1.1.2
# - added code so that in cases where the IP name cannot be resolved, 0.0.0.0
#   is written to the data file
#
# changes in version 1.1.1
# - added code to support ping on platforms other than Solaris, user
#   must chose by hand
#
# based on work done by Connie Logg and Les Cottrell at the Stanford Linear
# Accelerator Center

# this is required for the fixed close on SunOS and Solaris
# if you do not use these OSes, you can remove the require and change
# the POSIX::close back to a normal close
use POSIX;

use Fcntl;
require "ctime.pl";

$| = 1;

my $GLITE_LOCATION=$ENV{GLITE_LOCATION};
my $GLITE_LOCATION_LOG=$ENV{GLITE_LOCATION_LOG};
$dest_name = '';
my $list_parameter = '';
my $tool_cursor = 0;
my $list_cursor = 0;
my @list=();

#################### Customize me! ##############################

$ping_data_dir = "$GLITE_LOCATION/var/glite-pinger";

# full path of the ping command
$ping = "/bin/ping";

# IP name of this machine (fully qualified)
$src_name = `echo -n \`hostname -f\``;

# size of ping packets
$ping_size = $ARGV[1];

# maximum number of top-level forks allowed.
$max_process_count = 15;

# location of glite-netmon-rgma's communication FIFO
# FIXME: this should be something else in the end
$rgma_fifo = "$GLITE_LOCATION/var/glite-netmon2rgma/glite-netmon2rgma-fifo";

# Log file
$logfile = "$GLITE_LOCATION_LOG/e2emonit-pinger.log";

# Unique identifying name for the tool
$tool_name = "pinger";

###################################################################

# look up the IP address of this machine
($true_name,$aliases,$addrtype,$addrlength,@addrs) = gethostbyname($src_name);
if (!defined $true_name) {
  die "Message at: " . `date` . "...Can't resolve $src_name to IP address\n";
}
else {
  ($a, $b, $c, $d) = unpack('C4', $addrs[0]);
  $src_ip = "$a.$b.$c.$d";
}

###################################################################
# Scanning for argument / Get configfile(.conf)

my $self = $0;
$self =~  s!^.*/!!;
if (@ARGV != 2)
{
    die "Message at: " . `date` . "...Usage:  perl $self configfile(.conf) $ping_size\n ";
}

open (CONFIG_LIST, "<$ARGV[0]") or die "Message at: " . `date` . "...Can't open $ARGV[0]: $!\n";
@config_list_lines=<CONFIG_LIST>;
close(CONFIG_LIST);

###################################################################
# Configure parameters from e2emonit_configuration.txt

while ($_=shift(@config_list_lines)) {

        next if(/^\s$/);
        if(/^List/){
                $list_cursor = 1;
                next;
        }
        if (/^\#/){
                $list_cursor = 0;
                next;
        }
        if (/^Pinger/){
                $tool_cursor = 1;
                next;
        }
        chop();
        # Get monitoring list
        unless($list_cursor != 1){

                        push(@list,$_);

        }
        # Get parameters
        unless($tool_cursor != 1){

                # Check parameters format, get value
                if($_ =~ /^[\-]\D{1}\d*$/){

                        $list_parameter="$list_parameter $_";
                }
                last if(/^End/);
        }

}

# ping command
# for AIX, IRIX, LINUX, and Digital UNIX
$ping_cmd = "$ping -n $list_parameter -s \$ping_size \$dest_ip";


#Initialise the rgma producer. Ensure that we get both the 100 and the 1000 length pings in
&initialise_rgma(scalar(@list)*2,$true_name);

foreach $dest_name (@list) {

  undef $dest_ip, $send;

  $process_count++;

  if (!fork) {

    # this is required because of a bug in Perl 5 on SunOS and Solaris
    # that corrupts the position of the parent files
    POSIX::close(fileno(NODE_LIST));
    
    # this is the beginning of the child process

    undef $packet_loss;
    local($min, $avg, $max);

    $time = time;
    if (!defined $dest_ip) {
      ($true_name,$aliases,$addrtype,$addrlength,@addrs) = gethostbyname($dest_name);
      if (!defined $true_name) {
        &log_it($src_name, $src_ip, $dest_name, '0.0.0.0', $ping_size, $time);
	die "Message at: " . `date` . "...Can't resolve $dest_name to IP address\n";
      }
      else {
	($a, $b, $c, $d) = unpack('C4', $addrs[0]);
	$dest_ip = "$a.$b.$c.$d";
      }
    }
  
    # get timestamp
    $time= time;

    # send one ping to prime caches
    ($time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno, $pingtimes) = &ping_it($dest_ip, $ping_interval, $ping_size, 1);

    # if first ping was ok, send the bunch
    if (defined $packets_sent) {
      $ping_count=$send unless $send!~/\d/;
      ($time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno, $pingtimes) = &ping_it($dest_ip, $ping_interval, $ping_size, $ping_count);
    
    &log_it($src_name, $src_ip, $dest_name, $dest_ip, $ping_size, $time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno, $pingtimes);
 	
	if ($min && $max && $avg){
                $produce_all=1;
        }
        &produce_it($produce_all, $src_name, $src_ip, $dest_name, $dest_ip, $ping_size, $time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno, $pingtimes);

	} 
	exit;
  }

  else {
    # print "In parent process.\n";
   
    # wait if there are too many processes running
    while ($process_count >= $max_process_count) {
      $p = wait;
      #print "PID $p finished\n";
      $process_count--;
    }
  }
}
#close(NODE_LIST);

# wait for remaining processes to finish before we exit
while ($process_count > 0 ) {
  $p = wait;
  #print "PID $p finished\n";
  $process_count--;
}

# this subroutine logs entries to the ping record file
# it opens the file exclusively so that other processes will have to wait until
# this write is finished
sub log_it {
  my ($src_name, $src_ip, $dest_name, $dest_ip, $ping_size, $time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno_ref, $pingtimes_ref) = @_;
 
  # figure out which file to open (of the format ping-YYYY-MM.txt)
  my ($sec, $minutes, $hour, $mday, $month, $year, $wday, $yday, $isdst) = gmtime($time);
  $month++;
  $year += 1900;
  local($ping_data_fn) = sprintf "$ping_data_dir/ping-%4d-%2.2d\.txt", $year, $month;

  # try to open the file assuming it exists (this will be true most of
  # the time, so we try it first)
  if (!sysopen PINGDATA, $ping_data_fn, O_RDWR | O_EXCL | O_APPEND) {

    # file doesn't exist, so try creating it
    if (!(sysopen(PINGDATA, $ping_data_fn, (O_RDWR | O_EXCL | O_CREAT)))) {

      # file appeared sometime between the first and second sysopen calls,
      # try opening it again
      if (!sysopen PINGDATA, $ping_data_fn, O_RDWR | O_EXCL | O_APPEND){

	print STDERR "Can't open $ping_data_fn: $!\n";
	-1; # return error code
      }
    }
  }

  # first format the output string, then print it; avoid overlapping
  # IO with out subprocesses 

  my @seqno = @{$seqno_ref};
  my @pingtimes = @{$pingtimes_ref};

  my ($num) = $#seqno + 1;

  $output = "$src_name $src_ip $dest_name $dest_ip $ping_size $time $packets_sent $packets_rcvd";

  if ($packets_rcvd > 0) {
    $output .= " $min $avg $max ";
    $output .= join(' ', @seqno) . " " . join(' ', @pingtimes);
  }

  print PINGDATA "$output\n";

  close(PINGDATA);
} # end of log_it subroutine

# this subrouting initialises glite-netmon-rgma for us
sub initialise_rgma {
    my ($count,$host_name) = @_;

    # Don't try to write if it is not a fifo
    return if ! -p $rgma_fifo;

    chomp($count);
    chomp($host_name);

    # Give the server 30 seconds to accept our data, otherwise give up
    # the die will make the *eval* die, but not the current thread
    $SIG{ALRM} = sub{die "timeout"};

    eval {
	alarm(30);

	open(SERVER,"> $rgma_fifo") || return;
	print SERVER ("addmetric $tool_name rtt $count NMIdSource=$host_name tool=$tool_name\n");
	print SERVER ("addmetric $tool_name icmppacketloss $count NMIdSource=$host_name tool=$tool_name\n");
	close(SERVER);
	
	alarm(0);
    };

    $SIG{ALRM} = 'DEFAULT';
}

sub time2timestamp {
    my ($time)=@_;

    my ($seconds, $minutes, $hours, $day_of_month, $month, $year, $wday, $yday, $isdst) = gmtime($time);

    return sprintf("%04d%02d%02d%02d%02d%02dZ",$year+1900,$month+1,$day_of_month,$hours,$minutes,$seconds);
}

# this subroutine produces information into RGMA via glite-netmon-rgma
sub produce_it {

    my ($produce_all, $src_name, $src_ip, $dest_name, $dest_ip, $ping_size, $time, $packets_sent, $packets_rcvd, $min, $avg, $max, $seqno_ref, $pingtimes_ref) = @_;

    # Don't try to write if it is not a fifo
    return if ! -p $rgma_fifo;

    # Give the server 30 seconds to accept our data, otherwise give up
    # the die will make the *eval* die, but not the current thread
    $SIG{ALRM} = sub{die "timeout"};

    eval {
	alarm(30);

	open(SERVER,"> $rgma_fifo") || die "Message at: " . `date` . "...left eval, fifo not openend";
      	$time = &time2timestamp($time);
	
	if ($produce_all == 1){
                print SERVER ("measurement $tool_name rtt $src_name $dest_name pinger $ping_size $time $min $max $avg\n");
        }
        print SERVER ("measurement $tool_name icmppacketloss $src_name $dest_name pinger $ping_size $time ".(100*(1-$packets_rcvd/$packets_sent))."\n");


	close SERVER;
	
	alarm(0);
    };

    $SIG{ALRM} = 'DEFAULT';
}

# this subroutine sends the pings.  It relies on the UNIX ping program to
# do the dirty work
sub ping_it {
  my ($dest_ip, $ping_interval, $ping_size, $ping_count) = @_;
  my($time) = time;
  my($ping_output) ='' ;
  my($status);
  my($packets_sent, $packets_rcvd);
  my($min, $max, $avg);
  my(@seqno);
  my(@pingtimes);

  $_ = $ping_cmd;
  s/"/\\"/g;

  # This evaluates the variables left in the $ping_cmd string
  $eval_ping_cmd = eval qq/"$_"/;
  my $pid = open(PING, "-|");
  if ($pid == 0) {
    open (STDERR, ">&STDOUT") || die "Message at: " . `date` . "...Couldn't redirect stderr: $!\n";
    exec $eval_ping_cmd;
    $now_str = &ctime($time);
    print $now_str;
    die "$now_str: Ping execution failed for $dest_name ($dest_ip): $!";
  }

  while (<PING>) {
    $ping_output .= $_;

    # due to ping BUG the line for RH7.3 MUST be replaced   
    #if (/(\d+) packets transmitted, (\d+) packets received/) {
    # it would be good to fix it transparently for both distributions

    if (/(\d+) packets transmitted, (\d+) received/) {
      $packets_sent = $1;
      $packets_rcvd = $2;
    }
    elsif (/(\d+\.?\d*)\/(\d+\.?\d*)\/(\d+\.?\d*)/) {  #look for num/num/num
      $min = $1;
      $avg = $2;
      $max = $3;
    }
    elsif (/seq=(\d+).*time=(\d+\.?\d*) ms/) {
      # look for seq=NN and time=NN
      push(@seqno, $1);
      ($pingtime = $2) =~ s/\.$//; # delete trailing period, if needed
      push (@pingtimes, $pingtime); 
      }
  }

  # if packets_rcvd > packets_sent write to STDERR
  if($packets_rcvd>$packets_sent) {
    print STDERR "$ping_output\n";
  }

  # if close shows ping did not successfully complete and we
  # didn't get any ping results, dump the ping output to STDERR
    close PING;
    if ($? && !defined $packets_sent) {

      if (! open(OUTPUT,">>$logfile") ) {
                die "Message at: " . `date` . "...Can't open output file $logfile\n";
            }

            print OUTPUT "----------------------------------------------\n";
            print OUTPUT "Message at: " . `date`;
	    print OUTPUT "Ping of $dest_name ($dest_ip) failed:\n---\n$ping_output\n---\n";
	    print STDERR "Ping of $dest_name failed:\n---\n$ping_output\n---\n";
    }

  elsif (!defined $packets_sent) {
    print STDERR "Ping of $dest_name produced no information.\n";
  }

  # return results
  ($time, $packets_sent, $packets_rcvd, $min, $avg, $max, \@seqno, \@pingtimes);
}
